import 'dart:math';
import 'package:fis_common/logger/logger.dart';
import 'package:fis_jsonrpc/services/aIDiagnosis.m.dart';
import 'package:fis_measure/interfaces/date_types/point.dart';
import 'package:fis_measure/interfaces/process/urm/urm_data_processor.dart';
import 'package:fis_measure/process/workspace/rpc_bridge.dart';
import 'package:flutter/material.dart';
import 'package:vid/us/vid_us_probe.dart';
import '../application.dart';
import 'package:fis_measure/interfaces/process/workspace/point_info.dart';

/// URM 专业应用
class URMApplication extends Application {
  URMApplication(
    VidUsProbe probe, {
    required this.resultWidth,
    required this.resultHeight,
    required this.urmDataProcessor,
    this.onUpdateChart,
  }) : super(probe);

  IURMDataProcessor urmDataProcessor;

  int resultWidth; // 分析结果图像宽度

  int resultHeight; // 分析结果图像高度

  final ValueChanged<URMChartParams>? onUpdateChart;

  @override
  PointInfo createPointInfo(Offset offset, PointInfoType type) {
    final width = urmDataProcessor.showLeftRight
        ? displaySize.width / 2
        : displaySize.width;
    final height = displaySize.height;
    final x = offset.dx / width;
    final y = offset.dy / height;
    Rect? zoomArea = getZoomAreaInFullVisual();
    Offset percentOffset = Offset(x, y);
    if (zoomArea != null) {
      percentOffset = getZoomedPoint(zoomArea, percentOffset);
    }
    final info = PointInfo.fromOffset(percentOffset, type);
    info.hostVisualArea = currentVisualArea; // 未切换区域则沿用当前区域

    if (isAnnotationWorking) {
      activeAnnotationItem?.execute(info);
    } else {
      activeMeasureItem?.execute(info);
      if (type == PointInfoType.touchMove) {
        mobileTouchEvent.emit(this, offset); // 传出移动事件
      }
      if (type == PointInfoType.touchUp) {
        mobileTouchEndEvent.emit(this, offset); // 传出触摸结束事件
      }
    }
    return info;
  }

  // 获取Rect r 中的点 p 在全图的归一化坐标
  Offset getZoomedPoint(Rect r, Offset p) {
    double xPctInR = p.dx * r.width + r.left;
    double yPctInR = p.dy * r.height + r.top;
    return Offset(xPctInR, yPctInR);
  }

  @override
  double get displayScaleRatio {
    // return max(
    //     1,
    //     min(displaySize.width / resultWidth,
    //         displaySize.height / resultHeight));
    return 1.0;
  }

  void loadURMVisuals() {
    loadVisuals();
  }

  Future<URMMeasureProcessResult?> getURMMeasureResult({
    required URMMeasureType urmMeasureType,
    required URMROIType rOIType,
    List<UrmPoint>? srcDPoints,
    double? cMlength,
    double shellWidth = 0,
  }) async {
    try {
      print("调接口获取测量值 getURMMeasureResult");
      URMMeasureParams params = urmDataProcessor.getURMMeasureParams();
      URMMeasureProcessResult result =
          await RPCBridge.ins.rpc.aIDiagnosis.uRMMeasureProcessAsync(
        URMMeasureProcessRequest(
          remedicalCode: params.remedicalCode,
          token: RPCBridge.ins.userToken,
          urmMeasureType: urmMeasureType, // 手动传入
          rOIType: rOIType, // 手动传入
          srcDPoints: srcDPoints, // 手动传入
          cMlength: cMlength, // 手动传入
          shellWidth: shellWidth, // 手动传入
          phywidth: urmDataProcessor.phywidth, // 手动计算
          urmImageType: params.urmImageType, // 手动取值
          urmBlend: params.urmBlend, // 手动取值
          downsampleIndex: params.downsampleIndex, // 手动取值
          intPowerDen: params.intPowerDen, // 手动取值
          intPowerDir: params.intPowerDir, // 手动取值
          sigmaGauss: params.sigmaGauss, // 手动取值
          vessScale: params.vessScale, // 手动取值
          velMaxScaler: params.velMaxScaler, // 手动取值
          velMinScaler: params.velMinScaler, // 手动取值
          iterations: params.iterations, // 手动取值
          imgProcessVer: params.imgProcessVer, // 手动取值
          zoomOn: params.zoomOn, // 手动取值
          zoomRoix: params.zoomRoix, // 手动取值
          zoomRoiy: params.zoomRoiy, // 手动取值
          zoomRoiwidth: params.zoomRoiwidth, // 手动取值
          zoomRoiheight: params.zoomRoiheight, // 手动取值
          roix: params.roix, // 手动取值
          roiy: params.roiy, // 手动取值
          roiwidth: params.roiwidth, // 手动取值
          roiheight: params.roiheight, // 手动取值
          leftRight: params.leftRight, // 手动取值
          upDown: params.upDown, // 手动取值
          screenWidth: params.screenWidth, // 手动取值
          screenHeight: params.screenHeight, // 手动取值
          urmTraceDPoints: params.urmTraceDPoints, // 手动取值
        ),
      );
      return result;
    } catch (e) {
      logger.e("URM measure error($e)");
      print("getURMMeasureResult error: $e");
      return null;
    }
  }

  // 本地完整图像内归一化坐标转视图（接口中 screenWidth、screenHeight 指代区域）内归一化坐标
  // TODO 未完成正确的带zoom的转换
  UrmPoint localToView(DPoint point) {
    double px =
        point.x * urmDataProcessor.fullVisualPixelWidth; // a.width 是完整图像宽度
    double py =
        point.y * urmDataProcessor.fullVisualPixelHeight; // a.height 是完整图像高度
    double pxB = px - urmDataProcessor.screenX; // b.left 是视图区域的左边界像素
    double pyB = py - urmDataProcessor.screenY; // b.top 是视图区域的上边界像素
    double pctX = pxB / urmDataProcessor.screenWidth; // b.width 是视图区域的宽度像素
    double pctY = pyB / urmDataProcessor.screenHeight; // b.height 是视图区域的高度像素
    return UrmPoint(
      x: pctX,
      y: pctY,
    );
  }

  // 视图（接口中 screenWidth、screenHeight 指代区域）内归一化坐标转本地完整图像内归一化坐标
  DPoint viewToLocal(UrmPoint point) {
    double px = point.x * urmDataProcessor.screenWidth; // b.width 是视图区域的宽度像素
    double py = point.y * urmDataProcessor.screenHeight; // b.height 是视图区域的高度像素
    double pxB = px + urmDataProcessor.screenX; // b.left 是视图区域的左边界像素
    double pyB = py + urmDataProcessor.screenY; // b.top 是视图区域的上边界像素
    double pctX =
        pxB / urmDataProcessor.fullVisualPixelWidth; // a.width 是完整图像宽度
    double pctY =
        pyB / urmDataProcessor.fullVisualPixelHeight; // a.height 是完整图像高度
    return DPoint(
      pctX,
      pctY,
    );
  }

  // 视图（接口中 screenWidth、screenHeight 指代区域）内归一化矩形转本地完整图像内归一化矩形
  Rect viewToLocalRect(DRect point) {
    DPoint p1 = viewToLocal(UrmPoint(x: point.left, y: point.top));
    DPoint p2 = viewToLocal(UrmPoint(x: point.right, y: point.bottom));
    return Rect.fromLTRB(p1.x, p1.y, p2.x, p2.y);
  }

  Offset viewToLocalOffset(Offset offset) {
    double px = offset.dx * urmDataProcessor.screenWidth; // b.width 是视图区域的宽度像素
    double py =
        offset.dy * urmDataProcessor.screenHeight; // b.height 是视图区域的高度像素
    double pxB = px + urmDataProcessor.screenX; // b.left 是视图区域的左边界像素
    double pyB = py + urmDataProcessor.screenY; // b.top 是视图区域的上边界像素
    double pctX =
        pxB / urmDataProcessor.fullVisualPixelWidth; // a.width 是完整图像宽度
    double pctY =
        pyB / urmDataProcessor.fullVisualPixelHeight; // a.height 是完整图像高度
    return Offset(pctX, pctY);
  }

  // 类型转换
  List<UrmPoint> dPointsToUrmPoints(List<DPoint>? points) {
    if (points == null) return [];
    List<UrmPoint> urmPoints = [];
    for (var point in points) {
      urmPoints.add(localToView(point));
    }
    return urmPoints;
  }

  // 类型转换
  List<DPoint> urmPointsToDPoints(List<UrmPoint>? points) {
    if (points == null) return [];
    List<DPoint> urmPoints = [];
    for (var point in points) {
      urmPoints.add(viewToLocal(point));
    }
    return urmPoints;
  }

  // 获取zoom区域在完整视图下的归一化坐标
  Rect? getZoomAreaInFullVisual() {
    if (!urmDataProcessor.zoomOn) return null;
    final zoomCenter = urmDataProcessor.zoomCenterPct;
    final zoomCenterInFullVisual = viewToLocalOffset(zoomCenter);
    final scaleX = urmDataProcessor.zoomScaleX;
    final scaleY = urmDataProcessor.zoomScaleY;
    return Rect.fromCenter(
        center: zoomCenterInFullVisual, width: 1 / scaleX, height: 1 / scaleY);
  }
}

class URMChartParams {
  final List<Point> points;
  final double cmlength;
  final int maxPointIndex;
  final int minPointIndex;

  URMChartParams({
    required this.points,
    required this.cmlength,
    required this.maxPointIndex,
    required this.minPointIndex,
  });
}

class URMSimpleMeasureParams {
  final URMMeasureType urmMeasureType;
  final URMROIType rOIType;
  final List<UrmPoint>? srcDPoints;
  final double? cMlength;
  double? shellWidth = 0;

  URMSimpleMeasureParams({
    required this.urmMeasureType,
    required this.rOIType,
    this.srcDPoints,
    this.cMlength,
    this.shellWidth,
  });
}
